__author__ = 'Sergey Ragatsky'
__version__ = '0.0.4'

import re
import logging
from types import FunctionType
from itertools import product

logger = logging.getLogger(__name__)


def ittr(*args, **kwargs):
    """
    Decorator that adds iteration attributes to test method
    """
    def update_attr(func):
        # combining kwargs and args dict with ittr dict
        ittrs = {str(item): item for item in args or []}
        ittr_dict = getattr(func, 'ittr', dict())
        ittr_dict.update(ittrs)
        ittr_dict.update(kwargs)
        setattr(func, 'ittr', ittr_dict)

        return func
    return update_attr


class IttrMultiplier(type):
    """
    Multiples the tests in a given test class by
    the parameters given in ittr decorator.
    """
    def __new__(mcs, name, bases, dct):

        # check if the class been multiplied by IttrMultiplier
        if dct.get('is_multiplied'):
            return type.__new__(mcs, name, bases, dct)

        for attribute_name, attribute in list(dct.items()):
            # if not a method continue
            if not type(attribute) == FunctionType:
                logging.debug('attribute {0} is not a method'.format(attribute_name))
                continue

            # combine class decorator hook if exist with test method attributes
            try:
                attribute = ittr(**dct.get('__ittr__', {}))(attribute)
            except TypeError:
                logging.error('Illegal __ittr__ type, dict expected')

            # is method decorated with platform attr
            if not hasattr(attribute, 'ittr') or not attribute.ittr:
                logging.debug('method {0} has not attr decorator'.format(attribute_name))
                continue

            # create product of all the iterators
            b = [[(key, value) for value in values]
                 for key, values in attribute.ittr.items() if values]
            products = list(map(dict, product(*b)))

            for prod in products:
                logging.debug('method product: {0}'.format(prod))
                suffix = re.sub(r'\W+', '',
                                str(list(prod.values()))
                                .translate(None, "[]'")
                                .replace(',', '_'))
                logging.debug('method suffix: {0}'.format(suffix))

                # in case itts passed are empty
                if not suffix:
                    logging.debug('Empty suffix, product: {0}'.format(prod))
                    continue
                new_func_name = attribute_name + '_' + suffix

                # combine both product and ittr dict to be added to new method
                func_params = dict(attribute.__dict__, **prod)
                mirror_func = mcs._attribute_injector(attribute, **func_params)
                setattr(mirror_func, 'ittr', prod)

                # assign new name and docstring and save back at our class
                mirror_func.__name__ = new_func_name
                mirror_func.__doc__ = attribute.__doc__
                dct[new_func_name] = mirror_func

            # set no test flag to original test method
            attribute.__test__ = False

        # mark has been multiplied
        dct['is_multiplied'] = True
        return type.__new__(mcs, name, bases, dct)

    @classmethod
    def _attribute_injector(cls, func, **keywords):
        def injector(*fargs, **fkeywords):
            # transfer ittr and attr to self when called
            self = fargs[0]
            for name, value in keywords.items():
                setattr(self, name, value)
            return func(*fargs, **fkeywords)

        # transfers all attr and ittr to newfunc
        for name, value in keywords.items():
            setattr(injector, name, value)
        setattr(injector, 'keywords', keywords)
        return injector
